<div>
<h3>Quick links</h3>
<ul>
<li><a href="https://github.com/holgern/pyedflib"><img src="{{ pathto("_static/github.png", 1) }}" height="16" width="16" alt="" /> Fork on Github</a></li>
</ul>
</div>
