
# THIS FILE IS GENERATED FROM pyedflib SETUP.PY
short_version = '0.1.34'
version = '0.1.34'
full_version = '0.1.34'
git_revision = '1af9ced84921b304fecdc994e7e8b9a4c1674245'
release = True

if not release:
    version = full_version
